/*
Copyright 2017 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package changevalidator

import (
	"strings"
	"testing"
	"time"

	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/apiserver/pkg/admission"

	"github.com/kubernetes-incubator/service-catalog/pkg/apis/servicecatalog"
	scadmission "github.com/kubernetes-incubator/service-catalog/pkg/apiserver/admission"
	"github.com/kubernetes-incubator/service-catalog/pkg/client/clientset_generated/internalclientset"
	"github.com/kubernetes-incubator/service-catalog/pkg/client/clientset_generated/internalclientset/fake"
	informers "github.com/kubernetes-incubator/service-catalog/pkg/client/informers_generated/internalversion"
	core "k8s.io/client-go/testing"
)

// newHandlerForTest returns a configured handler for testing.
func newHandlerForTest(internalClient internalclientset.Interface) (admission.Interface, informers.SharedInformerFactory, error) {
	f := informers.NewSharedInformerFactory(internalClient, 5*time.Minute)
	handler, err := NewDenyPlanChangeIfNotUpdatable()
	if err != nil {
		return nil, f, err
	}
	pluginInitializer := scadmission.NewPluginInitializer(internalClient, f, nil, nil)
	pluginInitializer.Initialize(handler)
	err = admission.ValidateInitialization(handler)
	return handler, f, err
}

// newFakeServiceCatalogClientForTest creates a fake clientset that returns a
// ClusterServiceClassList with the given ClusterServiceClass as the single list item.
func newFakeServiceCatalogClientForTest(sc *servicecatalog.ClusterServiceClass) *fake.Clientset {
	fakeClient := &fake.Clientset{}

	scList := &servicecatalog.ClusterServiceClassList{
		ListMeta: metav1.ListMeta{
			ResourceVersion: "1",
		}}
	scList.Items = append(scList.Items, *sc)

	fakeClient.AddReactor("list", "clusterserviceclasses", func(action core.Action) (bool, runtime.Object, error) {
		return true, scList, nil
	})
	return fakeClient
}

// newServiceInstance returns a new instance for the specified namespace.
func newServiceInstance(namespace string, serviceClassName string, planName string) servicecatalog.ServiceInstance {
	instance := servicecatalog.ServiceInstance{
		ObjectMeta: metav1.ObjectMeta{Name: "instance", Namespace: namespace},
		Spec: servicecatalog.ServiceInstanceSpec{
			PlanReference: servicecatalog.PlanReference{
				ClusterServicePlanExternalName: planName,
			},
			ClusterServiceClassRef: &servicecatalog.ClusterObjectReference{
				Name: serviceClassName,
			},
		},
	}
	return instance
}

// newClusterServiceClass returns a new instance with the specified plan and
// UpdateablePlan attribute
func newClusterServiceClass(name string, plan string, updateablePlan bool) *servicecatalog.ClusterServiceClass {
	sc := &servicecatalog.ClusterServiceClass{
		ObjectMeta: metav1.ObjectMeta{Name: name},
		Spec: servicecatalog.ClusterServiceClassSpec{
			CommonServiceClassSpec: servicecatalog.CommonServiceClassSpec{
				PlanUpdatable: updateablePlan,
			},
		},
	}
	return sc
}

// setupInstanceLister creates a Service Instance and sets up a Instance Lister that
// retuns the instance
func setupInstanceLister(fakeClient *fake.Clientset) {
	instance := newServiceInstance("dummy", "foo", "original-plan-name")
	scList := &servicecatalog.ServiceInstanceList{
		ListMeta: metav1.ListMeta{
			ResourceVersion: "1",
		}}
	scList.Items = append(scList.Items, instance)

	fakeClient.AddReactor("list", "serviceinstances", func(action core.Action) (bool, runtime.Object, error) {
		return true, scList, nil
	})
}

// TestClusterServicePlanChangeBlockedByUpdateablePlanSetting tests that the
// Admission Controller will block a request to update an Instance's
// Service Plan
func TestClusterServicePlanChangeBlockedByUpdateablePlanSetting(t *testing.T) {
	sc := newClusterServiceClass("foo", "bar", false)
	fakeClient := newFakeServiceCatalogClientForTest(sc)
	handler, informerFactory, err := newHandlerForTest(fakeClient)
	if err != nil {
		t.Errorf("unexpected error initializing handler: %v", err)
	}
	setupInstanceLister(fakeClient)
	instance := newServiceInstance("dummy", "foo", "new-plan")
	informerFactory.Start(wait.NeverStop)
	err = handler.(admission.MutationInterface).Admit(admission.NewAttributesRecord(&instance, nil, servicecatalog.Kind("ServiceInstance").WithVersion("version"), instance.Namespace, instance.Name, servicecatalog.Resource("serviceinstances").WithVersion("version"), "", admission.Update, nil))
	if err != nil {
		if !strings.Contains(err.Error(), "The Service Class foo does not allow plan changes.") {
			t.Errorf("unexpected error %q returned from admission handler.", err.Error())
		}
	} else {
		t.Error("This should have been an error")
	}
}

// TestClusterServicePlanChangePermittedByUpdateablePlanSetting tests the
// Admission Controller verifying it allows an instance change to the
// plan name if the service class has specified PlanUpdatable=true
func TestClusterServicePlanChangePermittedByUpdateablePlanSetting(t *testing.T) {
	sc := newClusterServiceClass("foo", "bar", true)
	fakeClient := newFakeServiceCatalogClientForTest(sc)
	handler, informerFactory, err := newHandlerForTest(fakeClient)
	if err != nil {
		t.Errorf("unexpected error initializing handler: %v", err)
	}

	setupInstanceLister(fakeClient)

	instance := newServiceInstance("dummy", "foo", "new-plan")
	informerFactory.Start(wait.NeverStop)
	err = handler.(admission.MutationInterface).Admit(admission.NewAttributesRecord(&instance, nil, servicecatalog.Kind("ServiceInstance").WithVersion("version"), instance.Namespace, instance.Name, servicecatalog.Resource("serviceinstances").WithVersion("version"), "", admission.Update, nil))
	if err != nil {
		t.Errorf("Unexpected error: %v", err.Error())
	}
}
